/*
* DHCP offer montitor
* contact: rp [at] rpiz.com
*/
using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text;

namespace dhcp_check
{
    class udpc
    {
        public Socket s;
        public EndPoint dhcpsc;
        public int br = 0;
        public byte[] rbuf = new byte[1024];
    }

    class Program
    {
        static udpc uc = new udpc();

        static void check_packet(IAsyncResult ar)
        {
            Socket remote = (Socket)ar.AsyncState;

            if (remote.Connected)
            {
                try
                {
                    uc.br = remote.EndReceiveFrom(ar, ref uc.dhcpsc);

                    Console.WriteLine("check_packet: recv :: " + uc.br);
                    read_packet(uc.rbuf);
                    uc.s.BeginReceiveFrom(uc.rbuf, 0, 512, 0, ref uc.dhcpsc, new AsyncCallback(check_packet), uc.s);
                }
                catch
                {
                }
            }

        }

        static void read_packet(byte[] p)
        {
            BinaryReader r = new BinaryReader(new MemoryStream(p));
            string ip, srv = String.Empty, lease = String.Empty;
            byte otype = 0;

            r.ReadBytes(16);
            byte[] b = r.ReadBytes(4);
            
            // ip offered from the DHCP server
            ip = b[0] + "." + b[1] + "." + b[2] + "." + b[3];

            r.ReadBytes(220);

            while (r.BaseStream.Position < r.BaseStream.Length)
            {
                byte ocode;

                ocode = r.ReadByte();

                if (ocode != 0 && ocode != 255)
                {
                    byte olen;

                    olen = r.ReadByte();

                    switch (ocode)
                    {
                        // lease time
                        case 51:
                            {
                                lease = Convert.ToString((r.ReadByte() * Math.Pow(256, 3)) + (r.ReadByte() * Math.Pow(256, 2)) + (r.ReadByte() * 256) + r.ReadByte());
                                break;
                            }

                        // DHCP message type
                        case 53:
                            {
                                otype = r.ReadByte();
                                break;
                            }

                        // DHCP server identifier
                        case 54:
                            {
                                b = r.ReadBytes(4);
                                srv = b[0] + "." + b[1] + "." + b[2] + "." + b[3];
                                break;
                            }

                        default:
                            {
                                r.ReadBytes(olen);
                                break;
                            }
                    }
                }
            }

            if (otype == 2) Console.WriteLine("Added offer of " + ip + " from " + srv + " with lease time of " + lease + " seconds.");
            uc.br = 0;
        }

        static void Main(string[] args)
        {
            byte[] data = new byte[512];
            int bytes_sent = 0;
            bool no_opt = false;

            uc.s = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            uc.s.EnableBroadcast = true;
            uc.s.ExclusiveAddressUse = false;
            try
            {
                uc.s.SendTimeout = 5 * 1000;
                uc.s.ReceiveTimeout = 5 * 1000;
            }
            catch
            {
                // Solaris doesn't allow the UDP timeout options to be set
                // so we have to use asynchronous calls
                no_opt = true;
            }

            uc.dhcpsc = (EndPoint)new IPEndPoint(IPAddress.Any, 68);
            uc.s.Bind(uc.dhcpsc);

            IPEndPoint dhcpcs = new IPEndPoint(IPAddress.Broadcast, 67);

            // Generate a Transaction ID for this request
            byte[] xid = new byte[4];
            Random xid_r = new Random(8000);
            xid_r.NextBytes(xid);

            // Fake mac address
            // 00-13-D3-D5-27-73
            byte[] mac_a = new byte[16];
            mac_a[0] = 0;
            mac_a[1] = 19;
            mac_a[2] = 211;
            mac_a[3] = 213;
            mac_a[4] = 39;
            mac_a[5] = 115;

            // Actual DHCPDISCOVER packet
            byte[] dhcp_p = new byte[243];

            Array.Copy(xid, 0, dhcp_p, 4, 4);
            Array.Copy(mac_a, 2, dhcp_p, 28, 6);

            // Set the OP Code to BOOTREQUEST
            dhcp_p[0] = 1;
            // Set the Hardware Address Type to Ethernet
            dhcp_p[1] = 1;
            // Set the Hardware Address Length (number of bytes)
            dhcp_p[2] = 6;
            // Set the Broadcast Flag
            dhcp_p[10] = 128;
            // Set the Magic Cookie values
            dhcp_p[236] = 99;
            dhcp_p[237] = 130;
            dhcp_p[238] = 83;
            dhcp_p[239] = 99;
            // Set the DHCPDiscover Message Type Option
            dhcp_p[240] = 53;
            dhcp_p[241] = 1;
            dhcp_p[242] = 1;

            bytes_sent = uc.s.SendTo(dhcp_p, dhcpcs);
            Console.WriteLine("bytes sent: " + bytes_sent);

            bool timeout = false;
            DateTime stime = DateTime.Now;

            if (no_opt)
            {
                uc.s.BeginReceiveFrom(uc.rbuf, 0, 512, 0, ref uc.dhcpsc, new AsyncCallback(check_packet), uc.s);
            }

            while (!timeout)
            {
                if (!no_opt)
                {
                    try
                    {
                        uc.br = uc.s.ReceiveFrom(uc.rbuf, ref uc.dhcpsc);
                        Console.WriteLine("bytes rec: " + uc.br);
                    }
                    catch
                    {
                        timeout = true;
                    }

                    if (!timeout && uc.br > 0) read_packet(uc.rbuf);
                }
                else
                {
                    DateTime ntime = DateTime.Now;
                    TimeSpan elap = ntime - stime;

                    if (elap.Seconds > 6) timeout = true;
                }
            }

            try
            {
                uc.s.Shutdown(SocketShutdown.Both);
                uc.s.Close();
            }
            catch
            {
            }
        }
        
    }
}
